import zipfile
from pathlib import Path

import requests

from data_quality_check.utils.paths import get_data_path


def download_file(file_name: str, url: str) -> Path:
    data_path = get_data_path()
    data_path.mkdir(exist_ok=True)

    file_path = data_path / file_name

    if not file_path.exists():
        response = requests.get(url)
        file_path.write_bytes(response.content)

    return file_path


def download_zipped_file(file_name: str, url: str) -> Path:
    data_path = get_data_path()
    data_path.mkdir(exist_ok=True)

    file_path = data_path / file_name

    if not file_path.exists():
        response = requests.get(url)
        if response.status_code != 200:
            raise FileNotFoundError("Could not download resource")

        tmp_path = data_path / "tmp.zip"
        tmp_path.write_bytes(response.content)

        with zipfile.ZipFile(tmp_path, "r") as zip_file:
            zip_file.extract(file_path.name, data_path)

        tmp_path.unlink()

    return file_path.as_uri().__str__()



import re
import datetime as d
from datetime import datetime

__all__ = [
    'check_military_id', 'check_passport_id', 'check_residence_id', 'check_driver_id',
    'check_HongKong_and_Macao_permit_id',
    'check_national_id', 'check_national_id_detail',
    'check_mobile', 'check_mobile_is_reachable', 'check_car_plate_no',
    'get_gender_from_identity_no', 'get_age_from_identity_no', 'get_vehicle_age_of_months',
    'get_valid_prepaid_sales_dates',
    'get_valid_prepaid_dates_in_activity'
]


def check_regex_func(pattern):
    return lambda any_str: any_str if any_str and re.match(pattern, any_str) else None


area = {"11": "北京", "12": "天津", "13": "河北", "14": "山西", "15": "内蒙古", "21": "辽宁", "22": "吉林", "23": "黑龙江",
        "31": "上海", "32": "江苏", "33": "浙江", "34": "安徽", "35": "福建", "36": "江西", "37": "山东", "41": "河南", "42": "湖北",
        "43": "湖南", "44": "广东", "45": "广西", "46": "海南", "50": "重庆", "51": "四川", "52": "贵州", "53": "云南", "54": "西藏",
        "61": "陕西", "62": "甘肃", "63": "青海", "64": "宁夏", "65": "新疆", "71": "台湾", "81": "香港", "82": "澳门", "91": "国外"}


### 检验身份证号码的函数
def _check_national_id_detail(id_card):
    errors = ['身份证号码位数不对', '验证通过', '身份证号码出生日期超出范围或含有非法字符', '身份证号码校验错误', '身份证地区非法', '其他']
    id_card = str(id_card).upper()
    id_card = id_card.strip()
    id_card_list = list(id_card)

    try:
        # 15位身份号码检测
        if len(id_card) == 15:
            if id_card[0:2] not in area.keys():
                return 4, errors[4]
            try:
                datetime.strptime('19' + id_card[6:12], '%Y%m%d')
                return 1, errors[1]
            except ValueError:
                return 2, errors[2]

        # 18位身份号码检测
        elif len(id_card) == 18:
            # 地区校验
            if id_card[0:2] not in area.keys():
                return 4, errors[4]
            # 出生日期的合法性检查
            try:
                if datetime.strptime(id_card[6:14], '%Y%m%d'):
                    # //计算校验位
                    S = (int(id_card_list[0]) + int(id_card_list[10])) * 7 + \
                        (int(id_card_list[1]) + int(id_card_list[11])) * 9 + \
                        (int(id_card_list[2]) + int(id_card_list[12])) * 10 + \
                        (int(id_card_list[3]) + int(id_card_list[13])) * 5 + \
                        (int(id_card_list[4]) + int(id_card_list[14])) * 8 + \
                        (int(id_card_list[5]) + int(id_card_list[15])) * 4 + \
                        (int(id_card_list[6]) + int(id_card_list[16])) * 2 + \
                        int(id_card_list[7]) * 1 + int(id_card_list[8]) * 6 + int(id_card_list[9]) * 3
                    Y = S % 11
                    M = "F"
                    JYM = "10X98765432"
                    M = JYM[Y]  # 判断校验位
                    if M == id_card_list[17]:  # 检测ID的校验位
                        return 1, errors[1]
                    else:
                        return 3, errors[3]
            except ValueError:
                return 2, errors[2]
        else:
            return 0, errors[0]

    except Exception as e:
        return 5, errors[5] + f":{e.args[0]}"


# 检查居民身份证 id
def check_national_id(id_card):
    return id_card if _check_national_id_detail(id_card)[0] == 1 else None


def check_national_id_detail(id_card):
    return _check_national_id_detail(id_card)[1] if id_card else None


check_military_id = check_regex_func(r'^[\u4E00-\u9FA5](字第)([0-9a-zA-Z]{4,8})(号?)$')  # 检查军官证 id
check_passport_id = check_regex_func('^([a-zA-z]|[0-9]){9}$')  # 检查护照 id
check_residence_id = check_regex_func(r'^\d{9}$')  # 检查户口 id
check_driver_id = lambda id: check_regex_func(r'^\d{12}$')(id) if id and id[:2] in area else None  # 检查驾驶证 id
check_HongKong_and_Macao_permit_id = check_regex_func('^[HMhm]{1}([0-9]{10}|[0-9]{8})$')  # 检查港澳通行证 id
check_mobile = check_regex_func(r'^1[3-9]\d{9}$')
# 参考：https://cloud.tencent.com/developer/article/1361209
check_car_plate_no = check_regex_func(
    r'^(([京津沪渝冀豫云辽黑湘皖鲁新苏浙赣鄂桂甘晋蒙陕吉闽贵粤青藏川宁琼使领][A-Z](([0-9]{5}[DF])|([DF]([A-HJ-NP-Z0-9])[0-9]{4})))'
    r'|([京津沪渝冀豫云辽黑湘皖鲁新苏浙赣鄂桂甘晋蒙陕吉闽贵粤青藏川宁琼使领][A-Z][A-HJ-NP-Z0-9]{4}[A-HJ-NP-Z0-9挂学警港澳使领]))$')


def check_adms_reachable(phase_detail, lose_remark):
    import re
    error_msg = lambda check_field_msg, value: (
        f'检查 {check_field_msg} 通过: {value}', f'检查 {check_field_msg} 失败: {value}')

    class Check:
        def __init__(self, pass_condition: bool, check_field_name: str, check_field_value):
            self.is_passed = pass_condition
            self.msgs = error_msg(check_field_name, check_field_value)

        def msg(self):
            return self.msgs[0] if self.is_passed else self.msgs[1]

    reachable_state = ['跟进中已联系', '跟进中申请战败', '线索战败', '邀约成功未分配顾问', '邀约成功已分配顾问未到店', 'A卡', '订车申请', '订车',
                       '成交申请', '成交', '申请战败', '退订申请', '退订', '退车申请', '退车', 'A卡战败', ]
    empty_phone_remark_pattern = r'.*(空号|停机|假号|暂停服务|虚假|假的|无效号码|号码不存在|打不通|联系不上|无法接通|不接|关机|无人接听).*'

    reachable_phase = Check(phase_detail in reachable_state, 'ADMS phase_detail', phase_detail)
    reachable_remark = Check(not lose_remark or not (re.match(empty_phone_remark_pattern, lose_remark)),
                             'ADMS lose_remark', lose_remark)
    return reachable_phase.is_passed and reachable_remark.is_passed, '  &  '.join(
        [reachable_phase.msg(), reachable_remark.msg()])


from pyspark.sql.types import StructType, BooleanType, StringType, StructField, ArrayType
from pyspark.sql.functions import udf

check_mobile_is_reachable = udf(check_adms_reachable, StructType(
    [StructField('is_reachable', BooleanType()), StructField('detail', StringType())]))


def get_gender_from_identity_no(id_no):
    if check_national_id(id_no):
        if len(id_no) == 15 and id_no[14].isdigit() and int(id_no[14]) % 2 == 0:
            return '女'
        if len(id_no) == 15 and id_no[14].isdigit() and int(id_no[14]) % 2 == 1:
            return '男'
        if len(id_no) == 18 and id_no[16].isdigit() and int(id_no[16]) % 2 == 0:
            return '女'
        if len(id_no) == 18 and id_no[16].isdigit() and int(id_no[16]) % 2 == 1:
            return '男'
    return None


def get_age_from_identity_no(data_date, id_no) -> int:
    if check_national_id(id_no):
        birthdate_start_index_old = 6
        birthdate_end_index_old = 12
        birthdate_start_index = 6
        birthdate_end_index = 14
        if len(id_no) == 15:
            age = int(data_date[0:4]) - int('19' + id_no[birthdate_start_index_old: birthdate_start_index_old + 2])
            if data_date[4:] >= id_no[birthdate_start_index_old + 2: birthdate_end_index_old]:
                age += 1
            return age
        if len(id_no) == 18:
            age = int(data_date[0:4]) - int(id_no[birthdate_start_index: birthdate_start_index + 4])
            if data_date[4:] >= id_no[birthdate_start_index + 4: birthdate_end_index]:
                age += 1
            return age
    return None


def valid_prepaid_sales_dates(dates: []) -> []:
    if not dates:
        return []
    elif len(dates) == 0:
        return []
    elif len(dates) == 1:
        return dates
    else:
        dates = sorted(dates)
        valid_dates = [dates[0]]
        last_valid_date = dates[0]
        try:
            valid_boundary = (datetime.strptime(last_valid_date, "%Y-%m-%d")
                              + d.timedelta(days=30)).strftime("%Y-%m-%d")
            for date in dates[1:]:
                if (datetime.strptime(date, "%Y-%m-%d") - datetime.strptime(valid_boundary, "%Y-%m-%d")).days >= 0:
                    last_valid_date = date
                    valid_dates.append(last_valid_date)
                    valid_boundary = (datetime.strptime(last_valid_date, "%Y-%m-%d")
                                      + d.timedelta(days=30)).strftime("%Y-%m-%d")
                    continue
                else:
                    continue

        except Exception as e:
            raise e
        return valid_dates


get_valid_prepaid_sales_dates = udf(valid_prepaid_sales_dates, ArrayType(StringType()))


def check_if_new_prepaid(valid_dates: [], data_date: str) -> bool:
    if not valid_dates:
        return True
    elif len(valid_dates) == 0:
        return True
    for date in valid_dates:
        if (datetime.strptime(date, "%Y-%m-%d") - datetime.strptime(data_date, "%Y-%m-%d")).days < 0:
            return False
    return True


def valid_prepaid_dates_in_activity(valid_dates: [], start_date: str, duration: int) -> []:
    if not valid_dates:
        return []
    elif len(valid_dates) == 0:
        return []
    prepaid_dates_in_activity = []
    for date in valid_dates:
        if 0 <= (datetime.strptime(date, "%Y-%m-%d") - datetime.strptime(start_date, "%Y-%m-%d")).days <= duration:
            prepaid_dates_in_activity.append(date)
    return prepaid_dates_in_activity


def get_vehicle_age_of_months(sales_date_key: int, data_date: str):
    if sales_date_key is None:
        return None
    current_date = datetime.strptime(data_date, '%Y-%m-%d')
    sales_date = datetime.strptime(str(sales_date_key), '%Y%m%d')
    sales_date_previous = sales_date
    months_gap = (current_date.year - sales_date_previous.year) * 12 + current_date.month - sales_date_previous.month
    if current_date.day < sales_date_previous.day:
        months_gap -= 1
    return months_gap


get_valid_prepaid_dates_in_activity = udf(valid_prepaid_dates_in_activity, ArrayType(StringType()))



# from pandas_profiling.report.presentation.core.table import Table
# from pandas_profiling.report.presentation.flavours.html import templates
#
#
# class HTMLTable(Table):
#     def render(self) -> str:
#         return templates.
#
#         ("table.html").render(**self.content)



import unittest

from dataplat.base_test import LocalSpark
from dataplat.data_process.sql_processor import SqlProcessor
import udf
from udf import *
from udf import _check_national_id_detail


class UDFTest(unittest.TestCase):
    def test_check_mobile(self):
        self.assertIsNone(check_mobile(''))
        self.assertIsNone(check_mobile('123'))
        self.assertIsNone(check_mobile('0123456789j'))
        self.assertIsNone(check_mobile('01234567899'))
        self.assertIsNone(check_mobile('11234567899'))
        self.assertIsNone(check_mobile(None))
        self.assertEqual('13234567899', check_mobile('13234567899'))

    def test_check_id(self):
        ids = ['TEL15874001947', '000000000000000000', '000000000000', '', None]
        for id in ids:
            self.assertIsNone(udf.check_national_id(id))
            self.assertIsNone(udf.check_military_id(id))
            self.assertIsNone(udf.check_driver_id(id))
            self.assertIsNone(udf.check_HongKong_and_Macao_permit_id(id))
            self.assertIsNone(udf.check_passport_id(id))
            self.assertIsNone(udf.check_residence_id(id))

    def test_register_udf(self):
        spark = LocalSpark.get()
        SqlProcessor(spark, "").register_udfs_from_pyfile("./udf.py")

    def test_check_reachable(self):
        self.assertFalse(udf.check_adms_reachable(None, None)[0])
        self.assertFalse(udf.check_adms_reachable('跟进中已联系', '空号')[0])
        self.assertFalse(udf.check_adms_reachable('跟进中已联系', '空号！！！！')[0])
        self.assertFalse(udf.check_adms_reachable('跟进中已联系', '  空号！！！！')[0])

        self.assertFalse(udf.check_adms_reachable('跟进中已联系', ' 停机！！！ ')[0])
        self.assertFalse(udf.check_adms_reachable('跟进中已联系', ' 假号！！！ ')[0])
        self.assertFalse(udf.check_adms_reachable('跟进中已联系', ' 暂停服务！！！ ')[0])
        self.assertFalse(udf.check_adms_reachable('跟进中已联系', ' 虚假！！！ ')[0])
        self.assertFalse(udf.check_adms_reachable('跟进中已联系', ' 号码不存在！！！ ')[0])
        self.assertFalse(udf.check_adms_reachable('跟进中已联系', ' 假的！！！ ')[0])
        self.assertFalse(udf.check_adms_reachable('跟进中已联系', ' 无效号码！！！ ')[0])

        self.assertFalse(udf.check_adms_reachable('跟进中已联系', ' 联系不上！！！ ')[0])
        self.assertFalse(udf.check_adms_reachable('跟进中已联系', ' 无法接通！！！ ')[0])
        self.assertFalse(udf.check_adms_reachable('跟进中已联系', ' 不接！！！ ')[0])
        self.assertFalse(udf.check_adms_reachable('跟进中已联系', ' 关机！！！ ')[0])
        self.assertFalse(udf.check_adms_reachable('跟进中已联系', ' 无人接听！！！ ')[0])
        self.assertFalse(udf.check_adms_reachable('跟进中已联系', ' 打不通！！！ ')[0])

    def test_check_national_id_date_part_with_15_digits(self):
        dates = [
            ('950218', 1, '验证通过!'),
            ('960229', 1, '验证通过!'),
            ('000229', 2, '身份证号码出生日期超出范围或含有非法字符!'),
            ('961418', 2, '身份证号码出生日期超出范围或含有非法字符!'),
            ('951045', 2, '身份证号码出生日期超出范围或含有非法字符!')]
        for (date, code, msg) in dates:
            with self.subTest(msg):
                self.assertEqual(code, _check_national_id_detail('372525' + date + '000')[0])

    def test_check_national_id_date_part_with_18_digits(self):
        dates = [
            ('19960229', 1, '验证通过!'),
            ('19000229', 2, '身份证号码出生日期超出范围或含有非法字符!'),
            ('19941418', 2, '身份证号码出生日期超出范围或含有非法字符!'),
            ('19941045', 2, '身份证号码出生日期超出范围或含有非法字符!')]
        for (date, code, msg) in dates:
            with self.subTest(msg):
                self.assertEqual(code, _check_national_id_detail('372525' + date + '0000')[0])

    def test_check_national_id_date_part_and_check_bit_with_18_digits(self):
        self.assertEqual(1, _check_national_id_detail('372525199502180007')[0])
        self.assertEqual(3, _check_national_id_detail('372525199502180000')[0])  # 校验位错误：'身份证号码校验错误!'

    def test_check_car_plate_no(self):
        self.assertEqual('闽D12345', check_car_plate_no('闽D12345'))
        self.assertEqual('闽F12345', check_car_plate_no('闽F12345'))
        self.assertEqual(None, check_car_plate_no('闽F12345ab'))

    def test_get_gender_from_id_no(self):
        self.assertEqual('男', get_gender_from_identity_no('510781199402209411'))
        self.assertEqual('男', get_gender_from_identity_no('510781940220941'))
        self.assertEqual('女', get_gender_from_identity_no('110116198606040620'))
        self.assertEqual('女', get_gender_from_identity_no('110116860604062'))
        self.assertEqual(None, get_gender_from_identity_no('110116198606040621'))
        self.assertEqual(None, get_gender_from_identity_no('770116860604062'))
        self.assertEqual(None, get_gender_from_identity_no('1234567'))

    def test_get_age_from_id_no(self):
        self.assertEqual(28, get_age_from_identity_no('20210220', '510781199402209411'))
        self.assertEqual(28, get_age_from_identity_no('20210220', '510781940220941'))
        self.assertEqual(27, get_age_from_identity_no('20210219', '510781199402209411'))
        self.assertEqual(27, get_age_from_identity_no('20210219', '510781940220941'))
        self.assertEqual(27, get_age_from_identity_no('20200220', '510781199402209411'))
        self.assertEqual(27, get_age_from_identity_no('20200220', '510781940220941'))
        self.assertEqual(None, get_age_from_identity_no('20210220', '510781199402209412'))
        self.assertEqual(None, get_age_from_identity_no('20210220', '770116860604062'))

    def test_return_valid_prepaid_sales_dates(self):
        # 测试1：传入空串应返回空串
        case1, result1 = [], []

        # 测试2：list长度为1时原样返回
        case2, result2 = ['2021-01-01'], ['2021-01-01']

        # 测试3：排除30天内的无效订单
        case3, result3 = ['2021-01-01', '2021-01-30'], ['2021-01-01']

        # 测试4：不排除30天外的有效订单
        case4, result4 = ['2021-01-01', '2021-01-31'], ['2021-01-01', '2021-01-31']

        # 测试5：第3个订单距离上一个有效订单超过30天，距离上一个无效订单小于30天，应排除中间的无效订单，保留两个有效订单
        case5, result5 = ['2021-01-01', '2021-01-30', '2021-01-31'], ['2021-01-01', '2021-01-31']

        # 测试6：第3个订单距离上一个有效订单小于30天，距离上上个有效订单大于30天，应排除最后的订单，保留前两个有效订单
        case6, result6 = ['2021-01-01', '2021-03-01', '2021-03-05'], ['2021-01-01', '2021-03-01']

        # 测试7：第3个订单日期应该无效
        case7, result7 = ['2021-01-01', '2021-03-01', '2021-03-05', '2021-04-02'], ['2021-01-01', '2021-03-01', '2021-04-02']
        self.assertEqual(result1, udf.valid_prepaid_sales_dates(case1))
        self.assertEqual(result2, udf.valid_prepaid_sales_dates(case2))
        self.assertEqual(result3, udf.valid_prepaid_sales_dates(case3))
        self.assertEqual(result4, udf.valid_prepaid_sales_dates(case4))
        self.assertEqual(result5, udf.valid_prepaid_sales_dates(case5))
        self.assertEqual(result6, udf.valid_prepaid_sales_dates(case6))
        self.assertEqual(result7, udf.valid_prepaid_sales_dates(case7))

    def test_check_if_new_prepaid(self):
        # 测试1：如果没有历史订单记录，则视为新购
        case1, data_date1, result1 = [], '2021-01-01', True

        # 测试2：购买日期在活动开始之后，视为新购
        case2, data_date2, result2 = ['2021-01-01'], '2020-12-01', True

        # 测试3：购买日期在活动开始当天，视为新购
        case3, data_date3, result3 = ['2021-01-01'], '2021-01-01', True

        # 测试4：购买日期在活动开始之前，不视为新购
        case4, data_date4, result4 = ['2021-01-01'], '2021-01-15', False

        # 测试5：购买日期分布在在活动开始前后，不视为新购
        case5, data_date5, result5 = ['2021-01-01', '2021-01-31'], '2021-01-15', False
        self.assertEqual(result1, udf.check_if_new_prepaid(case1, data_date1))
        self.assertEqual(result2, udf.check_if_new_prepaid(case2, data_date2))
        self.assertEqual(result3, udf.check_if_new_prepaid(case3, data_date3))
        self.assertEqual(result4, udf.check_if_new_prepaid(case4, data_date4))
        self.assertEqual(result5, udf.check_if_new_prepaid(case5, data_date5))

    def test_check_valid_prepaid_dates_in_activity(self):
        # 测试1：
        case1, data_date1, duration1, result1 = [], '2021-01-01', 90, []
        case2, data_date2, duration2, result2 = ['2021-01-01'], '2020-12-01', 90, ['2021-01-01']
        case3, data_date3, duration3, result3 = ['2021-01-01'], '2021-01-01', 90, ['2021-01-01']
        case4, data_date4, duration4, result4 = ['2021-01-01'], '2021-01-15', 90, []
        case5, data_date5, duration5, result5 = ['2021-01-01', '2021-01-31', '2021-12-31'], '2021-01-15', 90, ['2021-01-31']
        self.assertEqual(result1, udf.valid_prepaid_dates_in_activity(case1, data_date1, duration1))
        self.assertEqual(result2, udf.valid_prepaid_dates_in_activity(case2, data_date2, duration2))
        self.assertEqual(result3, udf.valid_prepaid_dates_in_activity(case3, data_date3, duration3))
        self.assertEqual(result4, udf.valid_prepaid_dates_in_activity(case4, data_date4, duration4))
        self.assertEqual(result5, udf.valid_prepaid_dates_in_activity(case5, data_date5, duration5))

    def test_get_vehicle_age(self):
        v1_sales_data = '20210109'
        v2_sales_data = '20210110'
        v3_sales_data = '20210111'
        data_date = '2022-01-10'
        v1_age = get_vehicle_age_of_months(v1_sales_data, data_date)
        v2_age = get_vehicle_age_of_months(v2_sales_data, data_date)
        v3_age = get_vehicle_age_of_months(v3_sales_data, data_date)
        v4_age = get_vehicle_age_of_months(None, data_date)
        self.assertEqual(v1_age, 12)
        self.assertEqual(v2_age, 12)
        self.assertEqual(v3_age, 11)
        self.assertEqual(v4_age, None)


if __name__ == '__main__':
    unittest.main()



from pathlib import Path
from setuptools import find_packages, setup

# Read the contents of README file
source_root = Path(".")
with (source_root / "README.md").open(encoding="utf-8") as f:
    long_description = f.read()

# Read the requirements
with (source_root / "requirements.txt").open(encoding="utf8") as f:
    requirements = f.readlines()

version = "0.0.3"

with (source_root / "src" / "data_quality_check" / "version.py").open(
        "w", encoding="utf-8"
) as f:
    f.writelines(
        [
            '"""This file is auto-generated by setup.py, please do not alter."""\n',
            f'__version__ = "{version}"\n',
            "",
        ]
    )

setup(
    name='data-quality-check',
    version=version,
    author='Guanlong Zhou',
    author_email='markgzhou@outlook.com',
    description='Spark data quality check tool',
    packages=find_packages("src"),
    package_dir={"": "src"},
    python_requires=">=3.7",
    install_requires=requirements,
    classifiers=[
        'Development Status :: 3 - Alpha',
        # Choose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable"
        "License :: OSI Approved :: MIT License",
        "Environment :: Console",
        "Operating System :: OS Independent",
        'Intended Audience :: Developers',
        "Intended Audience :: Science/Research",
        'Topic :: Software Development :: Libraries :: Python Modules',
        "Operating System :: OS Independent",
        "Framework :: IPython",
        'Programming Language :: Python :: 3'
    ],
    keywords="spark data-quality-check python jupyter ipython",
    long_description=long_description,
    long_description_content_type="text/markdown",
    entry_points={
        'console_scripts': [
            'data-quality-check=data_quality_check.controller.console:main'
        ]
    },
)



from typing import Optional, List

from pydantic import BaseModel


# def _merge_dictionaries(dict1: dict, dict2: dict) -> dict:
#     """
#     Recursive merge dictionaries.
#
#     :param dict1: Base dictionary to merge.
#     :param dict2: Dictionary to merge on top of base dictionary.
#     :return: Merged dictionary
#     """
#     for key, val in dict1.items():
#         if isinstance(val, dict):
#             dict2_node = dict2.setdefault(key, {})
#             _merge_dictionaries(val, dict2_node)
#         else:
#             if key not in dict2:
#                 dict2[key] = val
#     return dict2


class ConfigMetadata(BaseModel):
    creator: str = 'ThoughtWorks Data Service'
    description: Optional[str] = None


class ConfigDataset(BaseModel):
    type: str = 'hive'
    name: str = None


class ConfigProfilingGeneral(BaseModel):
    columns: list = ['*']
    top_freq_values_number: int = 10


class ConfigProfilingCustomized(BaseModel):
    class CodeCheck(BaseModel):
        column: str
        codes: List[str]

    class KeyMappingCheck(BaseModel):
        column: str
        target_table: str
        target_column: str

    class PatternCheck(BaseModel):
        column: str
        pattern: str

    code_check: Optional[List[CodeCheck]]
    key_mapping_check: Optional[List[KeyMappingCheck]]
    pattern_check: Optional[List[PatternCheck]]


class ConfigProfiling(BaseModel):
    general: ConfigProfilingGeneral = ConfigProfilingGeneral()
    customized: Optional[ConfigProfilingCustomized] = None


class Config(BaseModel):
    title: str = "Spark Profiling Report"
    metadata: Optional[ConfigMetadata] = None
    dataset: ConfigDataset = None
    profiling: ConfigProfiling = None



from typing import Dict, List

from pyspark.sql import SparkSession
from data_quality_check.config import ConfigDataset, ConfigProfilingCustomized


class CodeCheckResult:
    field_name: str
    expected_codes: list
    null_row_count: int
    unexpected_unull_row_count: int
    unexpected_codes: list
    total_row_count: int

    def __init__(self, field_name: str, expected_codes: list, null_row_count: int = 0,
                 unexpected_unull_row_count: int = 0, unexpected_codes: list = [], total_row_count: int = 0):
        self.field_name = field_name
        self.expected_codes = expected_codes
        self.null_row_count = null_row_count
        self.unexpected_unull_row_count = unexpected_unull_row_count
        self.unexpected_codes = unexpected_codes
        self.total_row_count = total_row_count

    def __repr__(self):
        return str(self.to_dict())

    def to_dict(self):
        return {
            'field_name': self.field_name,
            'expected_codes': self.expected_codes,
            'unexpected_codes': self.unexpected_codes,
            'null_row_count': self.null_row_count,
            'unexpected_unull_row_count': self.unexpected_unull_row_count,
            'total_row_count': self.total_row_count
        }


class KeyMappingCheckResult:
    field_name: str
    target_table: str
    target_column: str
    outstanding_value_samples: list
    outstanding_row_count: int
    null_and_empty_row_count: int
    total_row_count: int

    def __init__(self, field_name: str, target_table: str, target_column: str,
                 outstanding_value_samples: list = [], outstanding_row_count: int = 0,
                 null_and_empty_row_count: int = 0,
                 total_row_count: int = 0):
        self.field_name = field_name
        self.target_table = target_table
        self.target_column = target_column
        self.outstanding_value_samples = outstanding_value_samples
        self.outstanding_row_count = outstanding_row_count
        self.null_and_empty_row_count = null_and_empty_row_count
        self.total_row_count = total_row_count

    def __repr__(self):
        return str(self.to_dict())

    def to_dict(self):
        return {
            'field_name': self.field_name,
            'target_table': self.target_table,
            'target_column': self.target_column,
            'outstanding_value_samples': self.outstanding_value_samples,
            'outstanding_row_count': self.outstanding_row_count,
            'null_and_empty_row_count': self.null_and_empty_row_count,
            'total_row_count': self.total_row_count
        }


class CustomizedProfilerResult:
    code_check_result: List[CodeCheckResult] = []
    key_mapping_check_result: List[KeyMappingCheckResult] = []

    def __init__(self, code_check_result: Dict = None, key_mapping_check_result: Dict = None):
        if code_check_result is not None:
            self.code_check_result = code_check_result

        if key_mapping_check_result is not None:
            self.key_mapping_check_result = key_mapping_check_result

    def has_code_check_result(self):
        return len(self.code_check_result) > 0

    def has_key_mapping_check_result(self):
        return len(self.key_mapping_check_result) > 0

    def to_dict(self):
        result = {}
        if self.has_code_check_result():
            result['code_check'] = []
            for cc in self.code_check_result:
                result['code_check'].append(cc.to_dict())

        # Key Mapping Check
        if self.has_key_mapping_check_result():
            result['key_mapping_check'] = []
            for kmc in self.key_mapping_check_result:
                result['key_mapping_check'].append(kmc.to_dict())
        return result


class CustomizedProfiler:
    dataset_config: ConfigDataset
    profiling_config: ConfigProfilingCustomized
    spark: SparkSession

    def __init__(self, spark, dataset_config: ConfigDataset,
                 customized_profiling_config: ConfigProfilingCustomized = ConfigProfilingCustomized()):
        self.spark = spark
        self.dataset_config = dataset_config
        self.profiling_config = customized_profiling_config

    def code_check(self, table_name: str, field_name: str, expected_codes: list = []):
        expected_codes.sort()
        expected_codes_raw_str = str(expected_codes)
        expected_codes_str = str(expected_codes_raw_str).replace('[', '').replace(']', '')

        sql = f'SELECT 1 FROM {table_name}  WHERE `{field_name}` IS NULL '
        null_row_count = self.spark.sql(sql).count()

        sql = f'SELECT 1 FROM {table_name} ' \
              f' WHERE `{field_name}` NOT IN ({expected_codes_str}) AND `{field_name}` IS NOT NULL'
        unexpected_unull_row_count = self.spark.sql(sql).count()

        sql = f'SELECT 1 FROM {table_name} '
        total_row_count = self.spark.sql(sql).count()

        sql = f'SELECT COLLECT_SET(DISTINCT(`{field_name}`)) AS result_set, 1 AS one FROM {table_name} ' \
              f' WHERE `{field_name}` NOT IN ({expected_codes_str}) GROUP BY one LIMIT 50'
        sql_result = self.spark.sql(sql).collect()
        if len(sql_result) > 0:
            unexpected_codes = sql_result[0]['result_set']
        else:
            unexpected_codes = []
        return CodeCheckResult(field_name, expected_codes=expected_codes, null_row_count=null_row_count,
                               unexpected_unull_row_count=unexpected_unull_row_count,
                               unexpected_codes=unexpected_codes, total_row_count=total_row_count)

    def key_mapping_check(self, table_name, field_name, target_table_name, target_field_name):
        sql = f'SELECT COLLECT_SET(result) AS result_set, 1 AS one from ' \
              f' (SELECT DISTINCT(aa.{field_name}) as result FROM {table_name} aa ' \
              f' LEFT ANTI JOIN {target_table_name} bb ' \
              f' ON (aa.{field_name} = bb.{target_field_name}) ' \
              f' WHERE TRIM(aa.{field_name})!=\'\' AND aa.{field_name} IS NOT NULL order by result asc LIMIT 100) ' \
              f' GROUP BY one'
        sql_result = self.spark.sql(sql).collect()
        if len(sql_result) > 0:
            outstanding_value_samples = sql_result[0]['result_set']
        else:
            outstanding_value_samples = []

        sql = f' SELECT 1 FROM {table_name} aa ' \
              f' LEFT ANTI JOIN {target_table_name} bb ' \
              f' ON (aa.{field_name} = bb.{target_field_name}) ' \
              f' WHERE TRIM(aa.{field_name})!=\'\' AND aa.{field_name} IS NOT NULL'
        outstanding_row_count = self.spark.sql(sql).count()

        sql = f'SELECT 1 FROM {table_name} '
        total_row_count = self.spark.sql(sql).count()

        sql = f' SELECT 1 FROM {table_name} WHERE TRIM({field_name})=\'\' OR {field_name} IS NULL'
        null_and_empty_row_count = self.spark.sql(sql).count()

        return KeyMappingCheckResult(field_name=field_name, target_table=target_table_name,
                                     target_column=target_field_name,
                                     outstanding_value_samples=outstanding_value_samples,
                                     outstanding_row_count=outstanding_row_count,
                                     null_and_empty_row_count=null_and_empty_row_count, total_row_count=total_row_count)

    def run(self, return_type='dict'):
        table_name = self.dataset_config.name
        customized_profiler_result = CustomizedProfilerResult()

        # Code Check
        code_check_list = self.profiling_config.code_check
        if code_check_list is not None and len(code_check_list) > 0:
            for cc in code_check_list:
                cc_result = self.code_check(table_name=table_name, field_name=cc.column, expected_codes=cc.codes)
                customized_profiler_result.code_check_result.append(cc_result)

        # Key Mapping Check
        key_mapping_check_list = self.profiling_config.key_mapping_check
        if key_mapping_check_list is not None and len(key_mapping_check_list) > 0:
            for kmc in key_mapping_check_list:
                kmc_result = self.key_mapping_check(table_name=table_name, field_name=kmc.column,
                                                    target_table_name=kmc.target_table,
                                                    target_field_name=kmc.target_column)
                customized_profiler_result.key_mapping_check_result.append(kmc_result)

        if return_type == 'dict':
            return customized_profiler_result.to_dict()
        else:
            return customized_profiler_result



from typing import List, Dict

from pyspark.sql import SparkSession, DataFrame
from pyspark.sql.utils import AnalysisException

from data_quality_check.config import ConfigProfilingGeneral, ConfigDataset


class GeneralFieldResult:
    field_name: str
    min_value = None
    max_value = None
    mean_value = None
    stddev_value = None
    zero_count = None  # 0
    null_count = None  # Null
    empty_count = None  # Empty String
    row_count = None
    duplicate_row_count = None
    unique_row_count = None
    value_count = None
    top_freq_count_by_values = None

    def __init__(self, field_name):
        self.field_name = field_name

    def __repr__(self):
        return str(self.to_dict())

    def to_dict(self):
        return {
            'field_name': self.field_name,
            'min_value': self.min_value,
            'max_value': self.max_value,
            'mean_value': self.mean_value,
            'stddev_value': self.stddev_value,
            'zero_count': self.zero_count,
            'null_count': self.null_count,
            'empty_count': self.empty_count,
            'row_count': self.row_count,
            'duplicate_row_count': self.duplicate_row_count,
            'unique_row_count': self.unique_row_count,
            'value_count': self.value_count,
            'top_freq_count_by_values': self.top_freq_count_by_values
        }

    def to_str_dict(self):
        return {
            'field_name': str(self.field_name),
            'min_value': str(self.min_value),
            'max_value': str(self.max_value),
            'mean_value': str(self.mean_value),
            'stddev_value': str(self.stddev_value),
            'zero_count': str(self.zero_count),
            'null_count': str(self.null_count),
            'empty_count': str(self.empty_count),
            'row_count': str(self.row_count),
            'duplicate_row_count': str(self.duplicate_row_count),
            'unique_row_count': str(self.unique_row_count),
            'value_count': str(self.value_count),
            'top_freq_count_by_values': str(self.top_freq_count_by_values)
        }


class GeneralProfilerResult:
    field_results: Dict[str, GeneralFieldResult] = {}

    def __init__(self, field_results: Dict = None):
        if field_results is None:
            self.field_results = {}
        else:
            self.field_results = field_results

    def to_dict(self):
        return dict((vi.field_name, vi.to_dict()) for vi in list(self.field_results.values()))


class GeneralProfiler:
    config: ConfigProfilingGeneral
    field_names: List[str]
    field_results: Dict
    spark: SparkSession
    general_profiler_result: GeneralProfilerResult = None

    def __init__(self, spark, df: DataFrame = None, config: ConfigProfilingGeneral = ConfigProfilingGeneral(),
                 dataset_config: ConfigDataset = None):
        if df is not None:
            self.df: DataFrame = df
        elif dataset_config is not None:
            select_items = ', '.join(config.columns)
            self.df = spark.sql(f'SELECT {select_items} FROM {dataset_config.name}')
        else:
            raise ValueError("Value missing error:"
                             " Must specify 1 of the following args: df or dataset_config.")

        self.spark = spark
        self.config: ConfigProfilingGeneral = config

    def get_field_summary_by_type(self, df, field_name: str, summary_type: str):
        try:
            result_df = df.select(field_name).selectExpr(f'{summary_type}(`{field_name}`) as value')
        except AnalysisException:
            return None
        return result_df.collect().pop()['value']

    def get_value_count(self, spark, df, field_name: str):
        df.createOrReplaceTempView('temp_profiling_tb')
        return spark.sql(f'SELECT 1 AS ct from temp_profiling_tb GROUP BY (`{field_name}`)').count()

    def get_top_freq_count_by_values(self, spark, df, field_name: str):
        df.createOrReplaceTempView('temp_profiling_tb')
        query = f'SELECT SUM(1) as ct, {field_name} as value from temp_profiling_tb GROUP BY (`{field_name}`) ' \
                f' ORDER BY ct DESC limit {self.config.top_freq_values_number}'
        rows = spark.sql(query).collect()
        return [{'value': row['value'], 'count': row['ct']} for row in rows]

    def get_unique_row_count(self, spark, df, field_name: str):
        df.createOrReplaceTempView('temp_profiling_tb')
        query = f'SELECT SUM(1) as ct, {field_name} as value from temp_profiling_tb GROUP BY ({field_name}) ' \
                f' HAVING ct=1'
        return spark.sql(query).count()

    def get_duplicated_row_count(self, spark, df, field_name: str):
        df.createOrReplaceTempView('temp_profiling_tb')
        query = f' SELECT COALESCE(SUM(ct),0) AS ct FROM ' \
                f'(' \
                f' SELECT SUM(1) as ct, {field_name} as value from temp_profiling_tb GROUP BY ({field_name}) ' \
                f' HAVING ct>1' \
                f')'
        return spark.sql(query).collect().pop()['ct']

    def get_count_by_field_condition(self, spark, df, field_condition: str):
        df.createOrReplaceTempView('temp_profiling_tb')
        query = f'SELECT 1 FROM temp_profiling_tb WHERE {field_condition}'
        return spark.sql(query).count()

    def run(self, return_type='dict'):
        df = self.df
        spark = self.spark

        if len(self.config.columns) < 1 or self.config.columns[0] == '*':
            self.field_names = self.df.columns
        else:
            self.field_names = self.config.columns

        if self.general_profiler_result is None:
            self.general_profiler_result = GeneralProfilerResult()
            field_results = self.general_profiler_result.field_results
            for field_name in self.field_names:
                field_result = GeneralFieldResult(field_name)
                field_result.row_count = self.get_field_summary_by_type(df, field_name, 'count')
                field_result.mean_value = self.get_field_summary_by_type(df, field_name, 'mean')
                field_result.stddev_value = self.get_field_summary_by_type(df, field_name, 'stddev')
                field_result.min_value = self.get_field_summary_by_type(df, field_name, 'min')
                field_result.max_value = self.get_field_summary_by_type(df, field_name, 'max')

                field_result.zero_count = self.get_count_by_field_condition(spark, df, f'`{field_name}` == 0')
                field_result.null_count = self.get_count_by_field_condition(spark, df, f'`{field_name}` is NULL')
                field_result.empty_count = self.get_count_by_field_condition(spark, df, f'`{field_name}` == ""')

                field_result.value_count = self.get_value_count(spark, df, field_name)
                field_result.top_freq_count_by_values = self.get_top_freq_count_by_values(spark, df, field_name)
                field_result.unique_row_count = self.get_unique_row_count(spark, df, field_name)
                field_result.duplicate_row_count = self.get_duplicated_row_count(spark, df, field_name)
                field_results[field_name] = field_result
        else:
            field_results = self.general_profiler_result.field_results

        if return_type == 'dataframe':
            data = [vi.to_str_dict() for vi in list(field_results.values())]
            return spark.createDataFrame(data).select('field_name', 'min_value', 'max_value', 'mean_value',
                                                      'stddev_value', 'zero_count', 'null_count',
                                                      'empty_count', 'row_count', 'duplicate_row_count',
                                                      'unique_row_count', 'value_count')
        elif return_type == 'dict':
            return self.general_profiler_result.to_dict()
        else:
            return self.general_profiler_result
